import type { IPosition } from '@difizen/libro-code-editor';
import type { LibroCodeCellView } from '@difizen/libro-codemirror-code-cell';
import type { BaseSearchProvider, SearchFilters, SearchMatch } from '@difizen/libro-search';
import type { Event } from '@difizen/mana-app';
import { DisposableCollection, Emitter } from '@difizen/mana-app';
import type { CodeMirrorSearchHighlighter } from './codemirror-search-highlighter.js';
import { CodeMirrorSearchHighlighterFactory } from './codemirror-search-protocol.js';
/**
 * Search provider for cells.
 */
export declare class CodemirrorCellSearchProvider implements BaseSearchProvider {
    protected toDispose: DisposableCollection;
    /**
     * CodeMirror search highlighter
     */
    protected cmHandler: CodeMirrorSearchHighlighter;
    /**
     * Current match index
     */
    protected currentIndex: number | undefined;
    /**
     * Current search filters
     */
    protected filters: SearchFilters | undefined;
    /**
     * Current search query
     */
    protected query: RegExp | null;
    protected stateChangedEmitter: Emitter<void>;
    protected _isActive: boolean;
    protected _isDisposed: boolean;
    protected lastReplacementPosition: IPosition | null;
    protected highlighterFactory: CodeMirrorSearchHighlighterFactory;
    protected cell: LibroCodeCellView;
    /**
     * Constructor
     *
     * @param cell Cell widget
     */
    constructor(highlighterFactory: CodeMirrorSearchHighlighterFactory, cell: LibroCodeCellView);
    /**
     * Get an initial query value if applicable so that it can be entered
     * into the search box as an initial query
     *
     * @returns Initial value used to populate the search box.
     */
    getInitialQuery(): string;
    disposed?: boolean | undefined;
    protected setEditor(): Promise<void>;
    /**
     * Changed signal to be emitted when search matches change.
     */
    get stateChanged(): Event<void>;
    /**
     * Current match index
     */
    get currentMatchIndex(): number | undefined;
    /**
     * Whether the cell search is active.
     *
     * This is used when applying search only on selected cells.
     */
    get isActive(): boolean;
    /**
     * Whether the search provider is disposed or not.
     */
    get isDisposed(): boolean;
    /**
     * Number of matches in the cell.
     */
    get matchesCount(): number;
    /**
     * Clear currently highlighted match
     */
    clearHighlight(): Promise<void>;
    /**
     * Dispose the search provider
     */
    dispose(): void;
    /**
     * Set `isActive` status.
     *
     * #### Notes
     * It will start or end the search
     *
     * @param v New value
     */
    setIsActive(v: boolean): Promise<void>;
    /**
     * Initialize the search using the provided options. Should update the UI
     * to highlight all matches and "select" the first match.
     *
     * @param query A RegExp to be use to perform the search
     * @param filters Filter parameters to pass to provider
     */
    startQuery(query: RegExp | null, filters?: SearchFilters): Promise<void>;
    /**
     * Stop the search and clean any UI elements.
     */
    endQuery(): Promise<void>;
    /**
     * Highlight the next match.
     *
     * @returns The next match if there is one.
     */
    highlightNext(): Promise<SearchMatch | undefined>;
    /**
     * Highlight the previous match.
     *
     * @returns The previous match if there is one.
     */
    highlightPrevious(): Promise<SearchMatch | undefined>;
    /**
     * Replace the currently selected match with the provided text.
     *
     * If no match is selected, it won't do anything.
     *
     * @param newText The replacement text.
     * @returns Whether a replace occurred.
     */
    replaceCurrentMatch(newText: string): Promise<boolean>;
    /**
     * Replace all matches in the cell source with the provided text
     *
     * @param newText The replacement text.
     * @returns Whether a replace occurred.
     */
    replaceAllMatches: (newText: string) => Promise<boolean>;
    /**
     * Get the current match if it exists.
     *
     * @returns The current match
     */
    protected getCurrentMatch(): SearchMatch | undefined;
    protected updateCodeMirror: () => Promise<void>;
}
//# sourceMappingURL=codemirror-cell-search-provider.d.ts.map